/***********************************************************************
 *
 * PROJECT: PalmXQ
 * FILE:    gen.c
 * AUTHOR:  Cheok Yan Cheng
 *
 * DESCRIPTION: Legal move generator.
 *
 * LAST REVISION:   24/02/2002
 *
 ***********************************************************************/

#include <PalmOS.h>
#include "palmxqrsc.h"
#include "data.h"
#include "protos.h"
#include "defs.h"

static void GenPush(ARMGLOBALS_PROTO UInt16 from, UInt16 to);

void Gen(ARMGLOBALS_PROTO_VOID)
{
    /* Counter */
    Int16 j;

    /* Piece */
    Int16 p;

    /* Location of a piece */
    Int16 from;

    /* MailBox90 */
    Int16 x;

    /* Destination of a piece */
    Int16 to;

    /* Cannon flag */
    Boolean fCannon;

    /* Can the piece slide */
    Boolean bSlide;
    
    /* So far, we have no moves for the current ply */
    FirstMove[Ply + 1] = FirstMove[Ply];

    for(from = 0; from < BRD_SIZE; from++)
    {
        if(Color[from] != Side)
        {
            continue;
        }

        p = Piece[from];

        for(j=0; j<8; j++)
        {
            // No more direction 
            if(!Offset[p][j])
            {
                break;
            }

            x = Mailbox90[from];
            fCannon = false;
            bSlide = (p==ROOK || p==CANNON);

            do
            {
                if (p==PAWN && Side==LIGHT)
                {
                    x -= Offset[p][j]; 
                }
                else
                {
                    x += Offset[p][j];
                }

                to = Mailbox182[x];

                if(to == -1 ||
                (LegalPosition[Side == DARK?to:89-to] & MaskPiece[p]) == 0)
                {
                    break;
                }

                if(!fCannon)
                {
                    // If no friend occupies my destination
                    if(Color[to] != Side)
                    {
                        switch(p)
                        {
                            case KNIGHT:
                                if(Color[from + KnightCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    GenPush(ARMGLOBALS from, to);
                                }
                                break;

                            case ELEPHANT:
                                if(Color[from + ElephantCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    GenPush(ARMGLOBALS from, to);
                                }
                                break;

                            case CANNON:
                                if(Color[to] == EMPTY)
                                {
                                    // Legal move(Not captured move)
                                    GenPush(ARMGLOBALS from, to);
                                }
                                break;

                            default:
                                // Legal move
                                GenPush(ARMGLOBALS from, to);

                        }   // switch(p)

                    }   // if(Color[to] != Side)

                    if(Color[to] != EMPTY)
                    {
                        if(p == CANNON)
                        {
                            fCannon = true;
                        }
                        else
                        {
                            break;
                        }
                    }

                }   // if(!fCannon)
                else
                {
                    if(Color[to] == XSide)
                    {
                        // Legal move
                        GenPush(ARMGLOBALS from, to);
                        break;
                    }

                    if(Color[to] != EMPTY)
                    {
                        break;
                    }
                }
            }

            while(bSlide);
        }   // j
    }   // from
}

void GenCap(ARMGLOBALS_PROTO_VOID)
{
    /* Counter */
    Int16 j;

    /* Piece */
    Int16 p;

    /* Location of a piece */
    Int16 from;

    /* MailBox90 */
    Int16 x;

    /* Destination of a piece */
    Int16 to;

    /* Cannon flag */
    Boolean fCannon;

    /* Can the piece slide */
    Boolean bSlide;
    
    /* So far, we have no moves for the current ply */
    FirstMove[Ply + 1] = FirstMove[Ply];

    for(from = 0; from < BRD_SIZE; from++)
    {
        if(Color[from] != Side)
        {
            continue;
        }

        p = Piece[from];

        for(j=0; j<8; j++)
        {
            // No more direction 
            if(!Offset[p][j])
            {
                break;
            }

            x = Mailbox90[from];
            fCannon = false;
            bSlide = (p==ROOK || p==CANNON);

            do
            {
                if (p==PAWN && Side==LIGHT)
                {
                    x -= Offset[p][j]; 
                }
                else
                {
                    x += Offset[p][j];
                }

                to = Mailbox182[x];

                if(to == -1 ||
                (LegalPosition[Side == DARK?to:89-to] & MaskPiece[p]) == 0)
                {
                    break;
                }

                if(!fCannon)
                {
                    // If are enemy occupies my destination
                    if(Color[to] == XSide)
                    {
                        switch(p)
                        {
                            case KNIGHT:
                                if(Color[from + KnightCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    GenPush(ARMGLOBALS from, to);
                                }
                                break;

                            case ELEPHANT:
                                if(Color[from + ElephantCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    GenPush(ARMGLOBALS from, to);
                                }
                                break;

                            case CANNON:
                                if(Color[to] == EMPTY)
                                {
                                }
                                break;

                            default:
                                // Legal move
                                GenPush(ARMGLOBALS from, to);

                        }   // switch(p)

                    }   // if(Color[to] == XSide)

                    if(Color[to] != EMPTY)
                    {
                        if(p == CANNON)
                        {
                            fCannon = true;
                        }
                        else
                        {
                            break;
                        }
                    }

                }   // if(!fCannon)
                else
                {
                    if(Color[to] == XSide)
                    {
                        // Legal move
                        GenPush(ARMGLOBALS from, to);
                        break;
                    }

                    if(Color[to] != EMPTY)
                    {
                        break;
                    }
                }
            }

            while(bSlide);
        }   // j
    }   // from
}

static void GenPush(ARMGLOBALS_PROTO UInt16 from, UInt16 to)
{
    GenT *gen;

    gen = &GenDat[FirstMove[Ply + 1]++];
    gen->m.b.from = from;
    gen->m.b.to = to;

    /* This data field will only be useful in the future implement of cchess
    engine. At this moment, we will just simply assign it to 0. */

    if(Color[to] != EMPTY)
    {
        gen->score = 1000 + (Int16)Piece[to] * 10 - (Int16)Piece[from];
    }
    else
    {
        gen->score = History[from][to];
    }
}

Boolean IsInCheck(ARMGLOBALS_PROTO UInt8 side)
{
    Int16 i;
    Int16 darkKingLocation;
    Int16 lightKingLocation;

    // Lookup table of king palace location
    for(i = 0; i < 9; i++)
    {
        darkKingLocation = KingPalaceDark[i];

        if(Piece[darkKingLocation] == KING)
        {
            break;
        }
    }

    for(i = 0; i < 9; i++)
    {
        lightKingLocation = KingPalaceLight[i];

        if(Piece[lightKingLocation] == KING)
        {
            break;
        }
    }

    // If "Fei Jiang" possible?
    if( (lightKingLocation-darkKingLocation) % BRD_WIDTH == 0 )
    {
        i = darkKingLocation;

        do
        {

            i += BRD_WIDTH;
        }while(Piece[i] == EMPTY);

        if(i == lightKingLocation)
        {
            return true;
        }
    }

    return IsInAttack(ARMGLOBALS side == DARK? darkKingLocation : lightKingLocation,
                    side ^ SIDE_MASK);
}

/***********************************************************************
 *
 * FUNCTION:    IsInAttack
 *
 * DESCRIPTION: Determines whether location square is attacked by side.
 *
 * PARAMETERS:  location    Location of the board to be determined.
 *              side        Side which attacks the board location.
 *
 * RETURNED:    True if the location is attacked by side. Else false.
 *
 * COMMENT:     For the cannon case, an empty location square is also
 *              considered being "attacked" by a cannon as long as the
 *              cannon is blocked with a piece.
 *
 ***********************************************************************/
Boolean IsInAttack(ARMGLOBALS_PROTO Int16 location, UInt8 side)
{
    /* Counter */
    Int16 j;

    /* Piece */
    Int16 p;

    /* Location of a piece */
    Int16 from;

    /* MailBox90 */
    Int16 x;

    /* Destination of a piece */
    Int16 to;

    /* Cannon flag */
    Boolean fCannon;

    /* Can the piece slide */
    Boolean bSlide;
    
    for(from = 0; from < BRD_SIZE; from++)
    {
        if(Color[from] != side)
        {
            continue;
        }

        p = Piece[from];

        for(j=0; j<8; j++)
        {
            // No more direction
            if(!Offset[p][j])
            {
                break;
            }

            x = Mailbox90[from];
            fCannon = false;
            bSlide = (p==ROOK || p==CANNON);

            do
            {
                if (p==PAWN && side==LIGHT)
                {
                    x -= Offset[p][j]; 
                }
                else
                {
                    x += Offset[p][j];
                }

                to = Mailbox182[x];

                if(to == -1 ||
                (LegalPosition[side == DARK?to:89-to] & MaskPiece[p]) == 0)
                {
                    break;
                }

                if(!fCannon)
                {
                    // If no friend occupies my destination
                    if(Color[to] != side)
                    {
                        switch(p)
                        {
                            case KNIGHT:
                                if(Color[from + KnightCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    if(location == to)
                                    {
                                        return true;
                                    }
                                }
                                break;

                            case ELEPHANT:
                                if(Color[from + ElephantCheck[j]] == EMPTY)
                                {
                                    // Legal move
                                    if(location == to)
                                    {
                                        return true;
                                    }
                                }
                                break;

                            case CANNON:
                                    // Legal move(Not captured move)
                                break;

                            default:
                                // Legal move
                                if(location == to)
                                {
                                    return true;
                                }

                        }   // switch(p)

                    }   // if(Color[to] != side)

                    if(Color[to] != EMPTY)
                    {
                        if(p == CANNON)
                        {
                            fCannon = true;
                        }
                        else
                        {
                            break;
                        }
                    }

                }   // if(!fCannon)
                else
                {
                    // Legal move
                    if(location == to)
                    {
                        return true;
                    }

                    if(Color[to] != EMPTY)
                    {
                        break;
                    }
                }
            }
            while(bSlide);

        }   // j
    }   // from

    return false;
}

Int16 Repeat(ARMGLOBALS_PROTO_VOID)
{
    Int16 i;
    Int16 b[BRD_SIZE];

    /* Count of squares that are different from the current position */
    Int16 c = 0;  

    /* Number of repetitions */
    Int16 r = 0;

    /* Is a repetition impossible? */
    if (Fifty <= 3)
    {
		return 0;
    }

    MemSet(b, sizeof(b), 0);

	/* loop through the reversible moves */
    /* for (i = HPly - 1; i >= HPly - Fifty - 1; --i) */
    for (i = HPly - 1; i >= HPly - Fifty; --i)
    {
        if (++b[HistDat[i].m.b.from] == 0)
			--c;
		else
			++c;
        if (--b[HistDat[i].m.b.to] == 0)
			--c;
		else
			++c;
		if (c == 0)
			++r;
	}

	return r;
}
